"""
Functions found in advapi32.dll

Security functions for manipulating the Windows Registry.
"""

import logging

from . import win_constants as wc
from ... import actions, utils, constants, objects
from ...call_hooks import builtin_func

logger = logging.getLogger(__name__)


def _get_reg_key(cpu_context, root_key_handle, sub_key_ptr, wide=False):
    """
    Retrieves the root key and sub key strings from the given hKey and lpSubKey arguments.

    :param root_key_handle: The hKey argument (the HKEY handle for the root key)
    :param sub_key_ptr: The lpSubKey argument (the pointer to the sub key string)
    :param wide: Whether strings are utf16 or utf8
    """
    # Make sure root_key_handle is 32-bit
    root_key_handle &= 0xFFFFFFFF
    # Get root key
    try:
        reg_key = cpu_context.objects[root_key_handle]
        root_key = reg_key.path
    except KeyError:
        if root_key_handle in wc.RegistryKey.__members__.values():
            # root key is a predefined enum.
            root_key = wc.RegistryKey(root_key_handle).name
        else:
            logger.warning("Invalid registry key 0x%X, using hex string.", root_key_handle)
            root_key = hex(root_key_handle)

    # Get sub key
    if not sub_key_ptr:
        sub_key = None
    else:
        sub_key = cpu_context.memory.read_string(sub_key_ptr, wide=wide)

    return root_key, sub_key


@builtin_func("RegOpenKeyA")
@builtin_func("RegOpenKeyW")
@builtin_func("RegOpenKeyExA")
@builtin_func("RegOpenKeyExW")
@builtin_func("RegCreateKeyA")
@builtin_func("RegCreateKeyW")
@builtin_func("RegCreateKeyExA")
@builtin_func("RegCreateKeyExW")
#typespec("LSTATUS RegOpenKeyA(HKEY hKey, LPCSTR lpSubKey, PHKEY phkResult)")
#typespec("int RegCreateKeyA(HKEY hKey, char* lpSubKey, HKEY* phkResult);")
#typespec("int RegCreateKeyW(HKEY hKey, wchar* lpSubKey, HKEY* phkResult);")
#typespec("int RegCreateKeyExA(HKEY hKey, char* lpSubKey, DWORD, char* lpClass, DWORD dwOptions, REGSAM samDesired, const SECURITY_ATTRIBUTES* lpSecurityAttributes, HKEY* phkResult, DWORD* lpdwDisposition);")
#typespec("int RegCreateKeyExA(HKEY hKey, wchar* lpSubKey, DWORD, wchar* lpClass, DWORD dwOptions, REGSAM samDesired, const SECURITY_ATTRIBUTES* lpSecurityAttributes, HKEY* phkResult, DWORD* lpdwDisposition);")
def reg_open_key(cpu_context, func_name, func_args):
    """
    Opens/Creates the specified registry key

    We are merging RegOpenKey* and RegCreateKey* because their effects are essentially same except
    for the extra disposition argument.
    """
    wide = func_name.endswith("W")
    root_key_handle, sub_key_ptr, *rest = func_args
    if func_name.startswith("RegCreateKeyEx"):
        result_ptr = rest[-2]  # RegCreateKeyEx has an extra disposition argument at the end.
    else:
        result_ptr = rest[-1]  # result pointer is last argument in both Ex and non-Ex versions.

    root_key, sub_key = _get_reg_key(cpu_context, root_key_handle, sub_key_ptr, wide)

    # If subkey is null or an empty string, result is the same as the root key handle.
    if not sub_key:
        disposition = wc.REG_OPENED_EXISTING_KEY
        cpu_context.memory.write(
            result_ptr, root_key_handle.to_bytes(cpu_context.byteness, cpu_context.byteorder)
        )
        logger.debug("Opening existing registry key: %s", root_key)

    # Otherwise create new RegKey object.
    else:
        disposition = wc.REG_CREATED_NEW_KEY
        path = "\\".join([root_key, sub_key])
        handle = cpu_context.objects.alloc()
        cpu_context.memory.write(
            result_ptr, handle.to_bytes(cpu_context.byteness, cpu_context.byteorder)
        )
        cpu_context.actions.add(
            actions.RegKeyOpened(cpu_context.ip, handle, path, root_key, sub_key)
        )
        logger.debug("Opening registry key: %s", path)

    # Need to report disposition if RegCreateKeyEx*
    if func_name.startswith("RegCreateKeyEx"):
        disposition_ptr = func_args[-1]
        if disposition_ptr:
            cpu_context.memory.write_data(disposition_ptr, disposition, data_type=constants.DWORD)

    return wc.ERROR_SUCCESS


@builtin_func("RegDeleteKeyA")
@builtin_func("RegDeleteKeyW")
@builtin_func("RegDeleteKeyExA")
@builtin_func("RegDeleteKeyExW")
#typespec("int RegDeleteKeyA(HKEY hKey, char* lpSubKey);")
#typespec("int RegDeleteKeyW(HKEY hKey, wchar* lpSubKey);")
#typespec("int RegDeleteKeyExA(HKEY hKey, char* lpSubKey, REGSAM samDesired, DWORD Reserved);")
#typespec("int RegDeleteKeyExW(HKEY hKey, wchar* lpSubKey, REGSAM samDesired, DWORD Reserved);")
def reg_delete_key(cpu_context, func_name, func_args):
    """
    Deletes a subkey and its values.
    """
    wide = func_name.endswith("W")
    root_key_handle, sub_key_ptr, *_ = func_args

    root_key, sub_key = _get_reg_key(cpu_context, root_key_handle, sub_key_ptr, wide)

    if not sub_key:
        logger.warning("Sub key is null")
        return

    path = "\\".join([root_key, sub_key])
    cpu_context.actions.add(actions.RegKeyDeleted(cpu_context.ip, root_key_handle, path))
    logger.debug("Deleting registry key: %s", path)

    return wc.ERROR_SUCCESS


@builtin_func("RegDeleteKeyValueA")
@builtin_func("RegDeleteKeyValueW")
#typespec("int RegDeleteKeyValueA(HKEY hKey, char* lpSubKey, char* lpValueName);")
#typespec("int RegDeleteKeyValueW(HKEY hKey, wchar* lpSubKey, wchar* lpValueName);")
def reg_delete_key_value(cpu_context, func_name, func_args):
    """
    Removes the specified value from the specified registry key and subkey.
    """
    wide = func_name.endswith("W")
    root_key_handle, sub_key_ptr, value_name_ptr = func_args

    root_key, sub_key = _get_reg_key(cpu_context, root_key_handle, sub_key_ptr, wide)
    value_name = cpu_context.memory.read_string(value_name_ptr, wide=wide)

    path = "\\".join([root_key, sub_key or ""])
    cpu_context.actions.add(
        actions.RegKeyValueDeleted(cpu_context.ip, root_key_handle, path, value_name)
    )
    logger.debug("Deleting value %s from registry key %s", value_name, path)

    return wc.ERROR_SUCCESS


@builtin_func("RegDeleteValueA")
@builtin_func("RegDeleteValueW")
#typespec("int RegDeleteValueA(HKEY hKey, char* lpValueName);")
#typespec("int RegDeleteValueW(HKEY hKey, wchar* lpValueName);")
def reg_delete_value(cpu_context, func_name, func_args):
    """
    Removes a named value from the specified registry key.
    """
    wide = func_name.endswith("W")
    root_key_handle, value_name_ptr = func_args

    root_key, _ = _get_reg_key(cpu_context, root_key_handle, 0, wide)
    value_name = cpu_context.memory.read_string(value_name_ptr, wide=wide)

    path = root_key
    cpu_context.actions.add(
        actions.RegKeyValueDeleted(cpu_context.ip, root_key_handle, path, value_name)
    )
    logger.debug("Deleting value %s from registry key %s", value_name, path)

    return wc.ERROR_SUCCESS


@builtin_func("RegSetValueA")
@builtin_func("RegSetValueW")
@builtin_func("RegSetValueExA")
@builtin_func("RegSetValueExW")
#typespec("int RegSetValueA(HKEY hKey, char* lpSubKey, DWORD dwType, char* lpData, DWORD cbData);")
#typespec("int RegSetValueA(HKEY hKey, wchar* lpSubKey, DWORD dwType, wchar* lpData, DWORD cbData);")
#typespec("int RegSetValueExA(HKEY, char* lpValueName, DWORD Reserved, DWORD dwType, const BYTE *lpData, DWORD cbData);")
#typespec("int RegSetValueExA(HKEY, wchar* lpValueName, DWORD Reserved, DWORD dwType, const BYTE *lpData, DWORD cbData);")
def reg_set_value(cpu_context, func_name, func_args):
    """
    Sets the data for the default or unnamed value of a specified registry key.
    """
    wide = func_name.endswith("W")
    root_key_handle, sub_key_ptr, *_, data_type, data_ptr, data_size = func_args

    root_key, sub_key = _get_reg_key(cpu_context, root_key_handle, sub_key_ptr, wide)

    path = "\\".join([root_key, sub_key or ""])
    data_type = wc.RegistryDataType(data_type)

    # Data is a null-terminated string
    if data_type.name in ("REG_SZ", "REG_LINK", "REG_EXPAND_SZ"):
        data = cpu_context.memory.read_string(data_ptr)

    # Data is a sequence of null-terminated strings, terminated by an empty string (\0)
    elif data_type.name == "REG_MULTI_SZ":
        # Since the last string is empty, this terminates as a double null,
        # which allows us to use WIDE_STRING.
        data = cpu_context.memory.read_data(data_ptr, data_type=constants.WIDE_STRING)
        data = [string.decode("utf8") for string in data.split(b"\x00") if string]

    elif data_type.name == "REG_BINARY":
        data = cpu_context.memory.read(data_ptr, data_size)

    elif data_type.name == "REG_DWORD":
        data = cpu_context.memory.read_data(data_ptr, data_type=constants.DWORD)

    elif data_type.name == "REG_QWORD":
        data = cpu_context.memory.read_data(data_ptr, data_type=constants.QWORD)

    elif data_type.name == "REG_NONE":
        data = None

    else:
        raise NotImplementedError(f"Unsupported data type: {data_type.name}")

    cpu_context.actions.add(
        actions.RegKeyValueSet(cpu_context.ip, root_key_handle, path, data_type.name, data)
    )
    logger.debug("Setting value %r to registry key %s", data, path)

    return wc.ERROR_SUCCESS


@builtin_func("OpenSCManagerA")
@builtin_func("OpenSCManagerW")
#typespec("HANDLE OpenSCManagerA(char* lpMachineName, char* lpDatabasename, DWORD dwDesiredAccess);")
#typespec("HANDLE OpenSCManagerA(wchar* lpMachineName, wchar* lpDatabasename, DWORD dwDesiredAccess);")
def open_sc_manager(cpu_context, func_name, func_args):
    """
    Establishes a connection to the service control manager on the specified computer
    and opens the specified service control manager database.
    """
    return cpu_context.objects.alloc()


@builtin_func("CreateServiceA")
@builtin_func("CreateServiceW")
#typespec("HANDLE CreateServiceA(HANDLE hSCManager, char* lpServiceName, char* lpDisplayName, DWORD dwDesiredAccess, DWORD dwServiceType, DWORD dwStartType, DWORD dwErrorControl, char* lpBinaryPathName, char* lpLoadOrderGroup, DWORD* lpdwTagId, char* lpDependencies, char* lpServiceStartName, char* lpPassword);")
#typespec("HANDLE CreateServiceW(HANDLE hSCManager, wchar* lpServiceName, wchar* lpDisplayName, DWORD dwDesiredAccess, DWORD dwServiceType, DWORD dwStartType, DWORD dwErrorControl, wchar* lpBinaryPathName, wchar* lpLoadOrderGroup, DWORD* lpdwTagId, wchar* lpDependencies, wchar* lpServiceStartName, wchar* lpPassword);")
def create_service(cpu_context, func_name, func_args):
    """
    Creates a service object and adds it to the specified service control manager database.
    """
    wide = func_name.endswith("W")
    service_name_ptr = func_args[1]
    display_name_ptr = func_args[2]
    desired_access = func_args[3]
    service_type = func_args[4]
    start_type = func_args[5]
    binary_path_ptr = func_args[7]

    service_name = cpu_context.memory.read_string(service_name_ptr, wide=wide)
    display_name = cpu_context.memory.read_string(display_name_ptr, wide=wide)
    binary_path = cpu_context.memory.read_string(binary_path_ptr, wide=wide)
    desired_access = wc.ServiceAccess(desired_access)
    service_type = wc.ServiceType(service_type)
    start_type = wc.ServiceStart(start_type)

    # TODO: Determine if a new handle is always created.
    handle = cpu_context.objects.get_or_alloc(objects.Service, name=service_name)

    action = actions.ServiceCreated(
        ip=cpu_context.ip,
        handle=handle,
        name=service_name,
        access=desired_access,
        service_type=service_type,
        start_type=start_type,
        display_name=display_name,
        binary_path=binary_path
    )
    cpu_context.actions.add(action)
    logger.debug("Created: %r", action)
    return handle


@builtin_func("OpenServiceA")
@builtin_func("OpenServiceW")
#typespec("SC_HANDLE OpenServiceA(SC_HANDLE hSCManager, LPCSTR lpServiceName, DWORD dwDesiredAccess);")
#typespec("SC_HANDLE OpenServiceW(SC_HANDLE hSCManager, LPCWSTR lpServiceName, DWORD dwDesiredAccess);")
def open_service(cpu_context, func_name, func_args):
    """
    Open a service
    """
    wide = func_name.endswith("W")
    service_name_ptr = func_args[1]

    service_name = cpu_context.memory.read_string(service_name_ptr, wide=wide)

    # TODO: Determine if a new handle is always created.
    handle = cpu_context.objects.get_or_alloc(objects.Service, name=service_name)

    action = actions.ServiceOpened(
        ip=cpu_context.ip,
        handle=handle,
        name=service_name
    )
    cpu_context.actions.add(action)
    logger.debug("Opened: %r", action)
    return handle


@builtin_func("DeleteService")
def delete_service(cpu_context, func_name, func_args):
    """
    BOOL DeleteService(SC_HANDLE hService);

    Delete service by handle
    """
    handle = func_args[0]

    if not handle:
        logger.warning(f"Service with handle {hex(handle)} does not exist.")
        
    try:
        action = actions.ServiceDeleted(
            ip=cpu_context.ip,
            handle=handle
        )
        cpu_context.actions.add(action)
        logger.debug("Deleted: %r", action)
    except ValueError as err:
        logger.debug("DeleteService: %s", err)

    return 1


@builtin_func("ChangeServiceConfig2A")
@builtin_func("ChangeServiceConfig2W")
#typespec("BOOL ChangeServiceConfig2A(SC_HANDLE hService, DWORD dwInfoLevel, LPVOID lpInfo);")
#typespec("BOOL ChangeServiceConfig2W(SC_HANDLE hService, DWORD dwInfoLevel, LPVOID lpInfo);")
def change_service_config(cpu_context, func_name, func_args):
    """
    Changes service configuration.

    NOTE: For now, only concerned with info level 1 for service description
    """
    wide = func_name.endswith("W")
    ptr_type = constants.QWORD if cpu_context.bitness == 64 else constants.DWORD
    handle = func_args[0]
    info_level = func_args[1]
    info_ptr = func_args[2]
    ptr_value = cpu_context.memory.read_data(info_ptr, data_type=ptr_type)

    if not handle:
        logger.warning(f"Service with handle {hex(handle)} does not exist.")

    if info_level == 1:  # SERVICE_CONFIG_DESCRIPTION
        service_description = cpu_context.memory.read_string(ptr_value, wide=wide)

        action = actions.ServiceDescriptionChanged(
            ip=cpu_context.ip,
            handle=handle,
            description=service_description
        )
        cpu_context.actions.add(action)
        logger.debug(f"Description for service with handle {hex(handle)} changed: {service_description}")

    return 1




from __future__ import annotations
from typing import TYPE_CHECKING, Tuple

from jpype import JImplements, JOverride

if TYPE_CHECKING:
    from docking.widgets.table import GTable
    from docking import ActionContext


@JImplements("dc3.rugosa.plugin.RugosaDockingAction")
class TableContextMenuEntry:
    NAME = ""

    def __init__(self, component: GTable):
        self._component = component

    def _get_coords(self) -> Tuple[int, int]:
        return self._component.getSelectedRow(), self._component.getSelectedColumn()

    def enabled(self, row: int, col: int) -> bool:
        return False

    @JOverride
    def isAddToPopup(self, context: ActionContext) -> bool:
        if context.getSourceComponent() != self._component:
            return False
        return self.enabled(*self._get_coords())

    def clicked(self, row: int, col: int):
        ...

    @JOverride
    def actionPerformed(self, context: ActionContext):
        return self.clicked(*self._get_coords())



"""
Interface for instruction management.
"""
import logging

from dragodis import BACKEND_GHIDRA
from dragodis.interface.instruction import ARMConditionCode

from .operands import ARMOperand
from ..exceptions import EmulationError
from ..instruction import Instruction

logger = logging.getLogger(__name__)


class ARMInstruction(Instruction):

    _operand_class = ARMOperand

    # Maps condition code to function to check if we can execute.
    _cond_map = {
        ARMConditionCode.EQ: lambda ctx: bool(ctx.registers.z),
        ARMConditionCode.NE: lambda ctx: bool(not ctx.registers.z),
        ARMConditionCode.CS: lambda ctx: bool(ctx.registers.c),
        ARMConditionCode.CC: lambda ctx: bool(not ctx.registers.c),
        ARMConditionCode.MI: lambda ctx: bool(ctx.registers.n),
        ARMConditionCode.PL: lambda ctx: bool(not ctx.registers.n),
        ARMConditionCode.VS: lambda ctx: bool(ctx.registers.v),
        ARMConditionCode.VC: lambda ctx: bool(not ctx.registers.v),
        ARMConditionCode.HI: lambda ctx: bool(ctx.registers.c and not ctx.registers.z),
        ARMConditionCode.LS: lambda ctx: bool(not ctx.registers.c or ctx.registers.z),
        ARMConditionCode.GE: lambda ctx: bool(
            (ctx.registers.n and ctx.registers.v)
            or (not ctx.registers.n and not ctx.registers.v)
        ),
        ARMConditionCode.LT: lambda ctx: bool(
            (ctx.registers.n and not ctx.registers.v)
            or (not ctx.registers.n and ctx.registers.v)
        ),
        ARMConditionCode.GT: lambda ctx: bool(
            not ctx.registers.z
            and (
                (ctx.registers.n and ctx.registers.v)
                or (not ctx.registers.n and not ctx.registers.v)
            )
        ),
        ARMConditionCode.LE: lambda ctx: bool(
            ctx.registers.z
            or (ctx.registers.n and not ctx.registers.v)
            or (not ctx.registers.n and ctx.registers.v)
        ),
        ARMConditionCode.AL: lambda ctx: True,
        ARMConditionCode.NV: lambda ctx: False,
    }

    @property
    def flag_update(self) -> bool:
        """
        Whether the condition flags are updated on the result of the operation.
        (S postfix)
        """
        return self._insn.update_flags

    def _check_condition(self) -> bool:
        """
        Checks condition flags to determine if instruction should be executed.
        """
        condition = self._insn.condition_code
        try:
            return self._cond_map[condition](self._cpu_context)
        except IndexError:
            raise EmulationError(f"Invalid condition code: {condition}")

    def _execute(self):
        # First check if conditions allow us to execute the instruction.
        if not self._check_condition():
            logger.debug("Skipping instruction at 0x%X. Condition code fails.", self.ip)
            return

        # Execute instruction.
        super()._execute()

        # If post-index or pre-index with update addressing mode,
        # update operand's base register based on offset.
        if self._insn.writeback:
            mnemonic = self._insn.mnemonic

            # Ignore push and pop instructions. We handle that in the opcode hook.
            if mnemonic in ("push", "pop"):
                return

            if "ldm" in mnemonic or "stm" in mnemonic:
                # write back for first operand (LDM/STM instruction)
                operand = self.operands[0]
                # In Ghidra, an stm*/ldm* instruction for push and pop will be treated as having
                # a single operand: sp!,{..}
                # This messes with our ability to update the writeback.
                # In the opcode hook we divert to a push/pop hook, but also need to be sure
                # we don't attempt the writeback here.
                if self._cpu_context.emulator.disassembler.name == BACKEND_GHIDRA and operand.text.startswith("sp!,"):
                    return
            else:
                # write back for last operand (! postfix or post-indexed operand)
                operand = self.operands[-1]

            logger.debug("Writeback operation for: %s", operand.text)
            operand.base += operand.offset



"""
Interface for handling the context of emulation: registers, memory, variables, actions, etc.
"""
from __future__ import annotations

import warnings
from copy import deepcopy
import collections
import logging
from typing import List, Tuple, Optional, TYPE_CHECKING

import dragodis
from dragodis import NotExistError

from . import utils
from .instruction import Instruction
from .memory import Memory
from .registers import RegisterMap
from .variables import VariableMap
from .operands import Operand, OperandLite
from .functions import FunctionSignature, FunctionArgument
from .objects import File, RegKey, Service, ObjectMap
from .actions import ActionList

if TYPE_CHECKING:
    from .emulator import Emulator

logger = logging.getLogger(__name__)


class JccContext:
    """
    Stores information pertaining to a Jcc instruction encountered when tracing.

    When a Jcc instruction is encountered, several pieces of information inherently need to be tracked since
    we are blindly taking every branch to ensure we get all possible data at any given address.  It turns out
    we need to know the target of the Jcc instruction for the condition as emulated
    (condition_target_ea).  We also need to know the value of the branch we would NOT have taken (at least as
    best of a guess as we can make in some cases) and where that value would have been set.  In order to
    calculate the value, we need to know what kind of test instruction was used, so that mnem is tracked as well.When
    we trace our condition_target_ea branch, we need not modify the context.  Whenever we trace the alternative branch,
    we'll need to modify the context as specified.
    """

    def __init__(self):
        self.condition_target_ea = None  # The branch actually taken
        self.alt_branch_data_dst = None  # The location which was tested (typically opnd 0 of the condition test)
        self.alt_branch_data = None  # The data stored in _alt_branc_data_dst
        self.flag_opnds = {}  # Dictionary containing the operands at a particular instruction which set
        # specific flags.  Dictionary is keyed on flag registery names.

    def __deepcopy__(self, memo):
        copy = JccContext()
        copy.condition_target_ea = self.condition_target_ea
        copy.alt_branch_data_dst = self.alt_branch_data_dst
        copy.alt_branch_data = self.alt_branch_data
        copy.flag_opnds = {flag: list(operands) for flag, operands in list(self.flag_opnds.items())}
        return copy

    def update_flag_opnds(self, flags, opnds):
        """
        Set the operands which last changed the specified flags.

        :param flags: list of flags which were modified utilizing the supplied opnds
        :param opnds: list of operands (instance of Operand) at the instruction which modified the flags
        """
        for flag in flags:
            # Converting Operand classes to OperandLite classes to help speed up deepcopies.
            self.flag_opnds[flag] = [OperandLite(opnd.ip, opnd.idx, opnd.text, opnd.value) for opnd in opnds]

    def get_flag_opnds(self, flags):
        """
        Extracts all the operands of for the list of flags and reduces the set.  However, since the operands
        need to remain in order, we can't use set operations.  In all actuality, assuming our code is correct and
        the compiler isn't doing something funky, any more than 1 flag should really just be a duplicate list.

        :param flags: list of flags for which to extract operands
        :return: list of operands which were utilized in the instruction that modified the requested flags
        """
        # TODO: Is there a better way to do this?
        operands = []
        for flag in flags:
            for operand in self.flag_opnds.get(flag, []):
                if operand not in operands:
                    operands.append(operand)

        return operands

    def is_alt_branch(self, ip):
        """
        Test our IP against the branch information to determine if we are in the branch that would have been
        emulated or in the alternate branch.
        """
        return self.condition_target_ea and self.condition_target_ea != ip


class ProcessorContext:
    """
    Stores the context of the processor during execution.

    :param emulator: Instance of Emulator to use during emulation.
    :param registers: Instance of an initialized RegisterMap object used to store register values
        for the given architecture.
    :param instruction_pointer: Name of the register used to point to the current instruction
        being currently executed or to-be executed.
    :param stack_pointer: Name of the register used to hold the stack pointer.
    :param return_register: Name of the register used to return results from a function.
    """

    # Must be set by inherited classes.
    OPCODES = {}  # Map of opcode mnemonics to functions that emulate them.

    # Class used to generate instructions.
    _instruction_class = Instruction

    # Cache for keeping track of instructions and their operand indexes.
    _operand_indices = {}

    def __init__(
            self,
            emulator: Emulator,
            registers: RegisterMap,
            instruction_pointer: str,
            stack_pointer: str,
            return_register: str
    ):
        self.emulator = emulator
        self.registers = registers
        self.jcccontext = JccContext()
        self.memory = Memory(self)
        self.call_history = []  # Keeps track of function calls.
        self.executed_instructions = []  # Keeps track of the instructions that have been executed.
        # TODO: Should memory_copies be stored in Memory object?
        self.memory_copies = collections.defaultdict(list)  # Keeps track of memory moves.
        self.bitness = emulator.disassembler.bit_size
        self.byteness = self.bitness // 8
        self.byteorder = "big" if emulator.disassembler.is_big_endian else "little"
        self.variables = VariableMap(self)
        self.objects = ObjectMap(self)
        self.actions = ActionList()
        self.stdout = ""  # stores contents of stdout.

        # Function start address of a function we are currently hooking.
        self.hooking_call = None
        self._call_depth = 0

        self._sp = stack_pointer
        self._ip = instruction_pointer
        self._ret = return_register
        self._sp_start = self.sp

    def __deepcopy__(self, memo):
        """Implementing our own deepcopy to improve speed."""
        # Create class, but avoid calling __init__()
        # so we don't trigger the unnecessary initialization of Memory and JccContext
        klass = self.__class__
        copy = klass.__new__(klass)
        memo[id(self)] = copy

        copy.emulator = self.emulator  # This is a reference, don't create a new instance.
        copy.hooking_call = self.hooking_call
        copy.registers = deepcopy(self.registers, memo)
        copy.jcccontext = deepcopy(self.jcccontext, memo)
        copy.memory = deepcopy(self.memory, memo)
        copy.variables = deepcopy(self.variables, memo)
        copy.objects = deepcopy(self.objects, memo)
        copy.actions = deepcopy(self.actions, memo)
        copy.call_history = list(self.call_history)
        copy.executed_instructions = list(self.executed_instructions)
        copy.memory_copies = self.memory_copies.copy()
        copy.stdout = self.stdout
        copy.bitness = self.bitness
        copy.byteness = self.byteness
        copy.byteorder = self.byteorder
        copy._call_depth = self._call_depth
        copy._sp = self._sp
        copy._ip = self._ip
        copy._ret = self._ret
        copy._sp_start = self._sp_start

        return copy

    @property
    def ip(self) -> int:
        """Alias for retrieving instruction pointer."""
        return self.registers[self._ip]

    @ip.setter
    def ip(self, value):
        """Alias for setting instruction pointer."""
        self.registers[self._ip] = value

    @property
    def sp(self) -> int:
        """Alias for retrieving stack pointer."""
        return self.registers[self._sp]

    @sp.setter
    def sp(self, value):
        """Alias for setting stack pointer."""
        self.registers[self._sp] = value

    @property
    def sp_diff(self) -> int:
        """
        The difference between the current stack pointer and the
        stack pointer at the beginning of the function.

        This helps with debugging since this number should match the number
        shown in the IDA disassembly.
        """
        return self._sp_start - self.sp

    # TODO: A subroutine in ARM can technically pass in larger values, in which
    #   case the value spans multiple registers r0-r3
    @property
    def ret(self) -> int:
        """Alias for retrieving the return value."""
        return self.registers[self._ret]

    @ret.setter
    def ret(self, value):
        """Alias for setting return value."""
        logger.debug("Setting 0x%X into %s", value, self._ret)
        self.registers[self._ret] = value

    @property
    def prev_instruction(self):
        """That last instruction that was executed or None if no instructions have been executed."""
        if self.executed_instructions:
            return self.executed_instructions[-1]
        else:
            return None

    def execute(self, start=None, end=None, call_depth: int = 0, max_instructions: int = None):
        """
        "Execute" the instruction at IP and store results in the context.
        The instruction pointer register will be set to the value supplied in .ip so that
        it is correct.

        :param start: instruction address to start execution (defaults to currently set ip)
        :param end: instruction to stop execution (not including)
            (defaults to only run start)
        :param call_depth: Number of function calls we are allowed to emulate into.
            When we hit our limit (depth is 0), emulation will no longer jump into function calls.
            (Defaults to not emulating into any function calls.)
            NOTE: This does not affect call hooks.
        :param max_instructions: Maximum number of instructions to execute before
            raising an RuntimeError.
            Uses max_instructions set by emulator constructor if not provided.

        :raises RuntimeError: If maximum number of instructions get hit.
        """
        if max_instructions is None:
            max_instructions = self.emulator.max_instructions

        if not start:
            start = self.ip

        # Set instruction pointer to where we are currently executing.
        self.ip = start

        # Set current call depth.
        if call_depth < 0:
            raise ValueError(f"call_depth must be a positive number.")
        self._call_depth = call_depth

        # If end is provided, recursively run execute() until ip is end.
        if end is not None:
            count = max_instructions
            while self.ip != end:
                instruction = self.instruction
                if instruction.is_terminal:
                    return  # TODO: Should we be executing the terminal instruction?
                instruction.execute()
                count -= 1
                if not count:
                    raise RuntimeError('Hit maximum number of instructions.')
            return
        else:
            self.instruction.execute()

    def _execute_call(self, func_address: int, func_name: str = None, call_address: int = None):
        """
        Executes the call to the given function.
        If a call hook has been provided, the hook will get run. Otherwise, instructions in the function
        will get emulated if set call_depth allows.

        :param func_address: Address of function to call.
        :param func_name: Name of the function (only required for name based call hook lookups)
        :param call_address: Address of call instruction (if available)
        """
        # Tell context that we are currently emulating a function hook.
        # This information is import for things like pulling out function arguments out correctly.
        self.hooking_call = func_address

        try:
            if not func_name:
                try:
                    func_name = self.emulator.disassembler.get_function_signature(func_address).name
                except NotExistError:
                    pass

            # See if we have a hook registered for this function.
            call_hook = self.emulator.get_call_hook(func_address)
            if not call_hook and func_name:
                call_hook = self.emulator.get_call_hook(func_name)
                if not call_hook:
                    # Try one more time with a sanitized name.
                    func_name = utils.sanitize_func_name(func_name)
                    call_hook = self.emulator.get_call_hook(func_name)

            # Report on function call and their arguments.
            arg_objs = self.get_function_args(func_address)
            self.call_history.append((call_address, func_name, [(arg.name, arg.value) for arg in arg_objs]))
            arg_values = [arg_obj.value for arg_obj in arg_objs]

            # Execute call hook if provided.
            if call_hook:
                try:
                    logger.debug(
                        "Emulating %s(%s)",
                        func_name,
                        ", ".join(f"{arg_obj.name}={hex(arg_obj.value)}" for arg_obj in arg_objs)
                    )
                    logger.debug("Running hook: %r", call_hook)
                    ret = call_hook(self, func_name, arg_values)
                    if ret is True:
                        ret = 1
                    elif ret is False:
                        ret = 0
                    # Set return value to rax
                    if ret is not None:
                        if not isinstance(ret, int):
                            raise TypeError(f"Invalid return type. Expected 'int' but got '{type(ret)}'")
                        self.ret = ret
                except RuntimeError:
                    raise  # Allow RuntimeError exceptions to be thrown.
                except Exception as e:
                    logger.debug("Failed to emulate builtin function: %s() with error: %s", func_name, e)

            # Otherwise, emulate the function instructions if set call_depth allows.
            elif self._call_depth:
                try:
                    func = self.emulator.disassembler.get_function(func_address)
                except NotExistError as e:
                    logger.warning("Failed to execute call at 0x%X : %s", func_address, e)
                    return
                call_depth = self._call_depth
                sp_start = self._sp_start
                try:
                    self.execute(start=func.start, end=func.end, call_depth=call_depth - 1)
                finally:
                    # Reset after we leave call.
                    self._call_depth = call_depth
                    self._sp_start = sp_start

        finally:
            self.hooking_call = None

    @property
    def func_calls(self) -> dict:
        warnings.warn(f".func_calls is deprecated. Please use .call_history instead.", DeprecationWarning)
        # Original .func_calls overwrote calls at the same instruction. So pulling the last one to replicate this logic.
        return {
            address: (func_name, [value for name, value in args])
            for address, func_name, args in self.call_history
        }

    def get_call_history(self, func_name_or_ea) -> List[Tuple[int, List]]:
        """
        Returns the call history for a specific function name.

        :returns: List of tuples containing: (ea of call, list of function arguments)
        """
        if isinstance(func_name_or_ea, str):
            func_name = func_name_or_ea
        else:
            ea = func_name_or_ea
            func_name = self.emulator.disassembler.get_function_signature(ea).name
        return [
            (address, [value for name, value in args])
            for address, _func_name, args in self.call_history
            if _func_name == func_name
        ]

    def prep_for_branch(self, bb_start_ea):
        """
        Modify this current context in preparation for a specific path.
        """
        if self.jcccontext.is_alt_branch(bb_start_ea):
            logger.debug("Modifying context for branch at 0x%08X", bb_start_ea)
            # Set the destination operand relative to the current context
            # to a valid value that makes this branch true.
            dst_opnd = self.jcccontext.alt_branch_data_dst
            dst_opnd = self.get_operands(ip=dst_opnd.ip)[dst_opnd.idx]
            dst_opnd.value = self.jcccontext.alt_branch_data

        self.jcccontext = JccContext()

    def get_instruction(self, ip=None) -> Instruction:
        """
        Gets the Instruction object for the current instruction pointed by the instruction pointer.

        :param ip: location of instruction pointer to pull Instruction from (default to current ip in context)
        :return: Instruction object
        """
        if ip is None:
            ip = self.ip
        return self._instruction_class(self, ip)

    @property
    def instruction(self) -> Instruction:
        return self.get_instruction()

    def get_operands(self, ip=None) -> List[Operand]:
        """
        Gets the Operand objects of all operands in the current instruction and returns them in a list.

        :param int ip: location of instruction pointer to pull operands from (defaults to current rip in context)

        :return: list of Operand objects
        """
        return self.get_instruction(ip=ip).operands

    @property
    def operands(self) -> List[Operand]:
        return self.get_operands()

    def get_pointer_history(self, ea):
        """
        Retrieves the history of a specific pointer.
        :param ea: Pointer to start with.
        :return: list of tuples containing (address of the memory copy, source pointer)
            - sorted by earliest to latest incarnation of the pointer. (not including itself)
        """
        history = []
        for ip, copies in sorted(list(self.memory_copies.items()), reverse=True):
            for src, dst, size in sorted(copies, reverse=True):
                if dst == ea:
                    history.append((ip, src))
                    ea = src
        history.reverse()
        return history

    def get_original_location(self, addr):
        """
        Retrieves the original location for a given address by looking through it's pointer history.

        :param addr: address of interest

        :return: a tuple containing:
            - instruction pointer where the original location was first copied
                or None if given address is already loaded or the original location could not be found.
            - either a loaded address, a tuple containing (frame_id, stack_offset) for a stack variable,
                or None if the original location could not be found.
        """
        # TODO: Consider refactoring.

        # Pull either the first seen loaded address or last seen stack variable.
        try:
            line = self.emulator.disassembler.get_line(addr)
            if not line.is_loaded:
                return None, addr
        except dragodis.NotExistError:
            return None, addr

        ip = None

        var = self.variables.get(addr, None)
        for ip, ea in reversed(self.get_pointer_history(addr)):
            try:
                line = self.emulator.disassembler.get_line(ea)
                if line.is_loaded:
                    return ip, ea
            except dragodis.NotExistError:
                pass
            var = self.variables.get(ea, var)

        if var and var.is_stack:
            # TODO: frame_id is an IDA specific thing. We are going to need to refactor this.
            #   - Perhaps have a "Variable" or "Data" object in dragodis?
            return ip, (var.frame_id, var.stack_offset)
        else:
            return ip, None

    def get_function_signature(
            self, func_ea: int = None, num_args: int = None, default_data_type: str = None
    ) -> Optional[FunctionSignature]:
        """
        Returns the function signature of the given func_ea with argument values pulled
        from this context.

        :param int func_ea: address of the function to pull signature from.
            The first operand is used if not provided. (helpful for a "call" instruction)
        :param int num_args: Force a specific number of arguments in the signature.
            If not provided, number of arguments is determined by the disassembler.
            Extra arguments not defined by the disassembler are assumed to be the default_data_type.
            Avoid using num_args and adjust the returned FunctionSignature manually
            if more customization is needed.

            WARNING: Setting the number of arguments will permanently change the
            signature on the backend disassembler.
        :param str default_data_type: The default data type to use when forcing extra arguments.
            (Defaults to "dword" for 32-bit or "qword" for 64-bit)

        :return: FunctionSignature object or None if not applicable

        :raises NotExistError: If a function signature could not be created from given ea.
        :raises ValueError: If num_args is negative
        """
        # If func_ea is not given, assume we are using the first operand from a call instruction.
        if not func_ea:
            if not self.operands:
                return None
            operand = self.operands[0]
            # function pointer can be a memory reference or immediate.
            func_ea = operand.addr or operand.value

        disassembler = self.emulator.disassembler
        signature = disassembler.get_function_signature(func_ea)
        signature = FunctionSignature(self, func_ea, signature)

        if num_args is not None:
            if num_args < 0:
                raise ValueError("num_args is negative")

            arguments = signature.arguments
            if len(arguments) > num_args:
                for _ in range(len(arguments) - num_args):
                    signature.remove_argument(-1)

            elif len(arguments) < num_args:
                if not default_data_type:
                    default_data_type = "qword" if self.bitness == 64 else "dword"
                for _ in range(num_args - len(arguments)):
                    signature.add_argument(default_data_type)

        return signature

    def get_function_args(
            self, func_ea: int = None, num_args: int = None, default_data_type: str = None
    ) -> List[FunctionArgument]:
        """
        Returns the FunctionArg objects for this context based on the
        given function.

        >>> cpu_context = ProcessorContext()
        >>> args = cpu_context.get_function_args(0x180011772)

        :param int func_ea: Ea of the function to pull a signature from.
        :param int num_args: Force a specific number of arguments.
            If not provided, number of arguments is determined by the disassembler.
            Extra arguments not defined by the disassembler are assumed to be 'int' type.
            Use get_function_signature() and adjust the FunctionSignature manually
            if more customization is needed.
            (NOTE: The function signature will be forced on failure if this is set.)
        :param str default_data_type: The default data type to use when forcing extra arguments.
            (Defaults to "dword" for 32-bit or "qword" for 64-bit)

        :returns: list of FunctionArg objects
        """
        func_sig = self.get_function_signature(func_ea, num_args=num_args, default_data_type=default_data_type)
        if not func_sig:
            return []

        return func_sig.arguments

    def get_function_arg_values(
            self, func_ea: int = None, num_args: int = None, default_data_type: str = None
    ) -> List[int]:
        """
        Returns the FunctionArg values for this context based on the given function.
        """
        return [arg.value for arg in self.get_function_args(func_ea=func_ea, num_args=num_args, default_data_type=default_data_type)]

    @property
    def function_args(self) -> List[FunctionArgument]:
        """
        The function arguments currently set based on the function in the first operand of the current instruction.
        """
        return self.get_function_args()

    @property
    def function_arg_values(self) -> List[int]:
        """
        The function argument values currently set based on the function in the first operand of the current instruction.
        """
        return self.get_function_arg_values()

    @property
    def passed_in_args(self) -> List[FunctionArgument]:
        """
        The function arguments for the current function.
        """
        func = self.emulator.disassembler.get_function(self.ip)
        return self.get_function_args(func.start)

    @property
    def files(self) -> List[File]:
        """
        All File objects in the current context.
        """
        return list(self.objects.query(File))

    @property
    def reg_keys(self) -> List[RegKey]:
        """
        The opened registry keys for this context.
        """
        return list(self.objects.query(RegKey))

    @property
    def services(self) -> List[Service]:
        """
        The created services for this context.
        """
        return list(self.objects.query(Service))



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Tuple

import ida_kernwin
from hexdump import hexdump
from PyQt5 import QtWidgets

if TYPE_CHECKING:
    from rugosa.ida_plugin.components.emulator import EmulatorForm


class Memory:

    def __init__(self, ui: EmulatorForm):
        self._ctx = None
        self.ui = ui
        self.start = ui.memory_start
        self.size = ui.memory_size
        self.hexdump = ui.memory_hexdump
        self.blocks_table = ui.memory_blocks_table
        self.blocks_table.itemDoubleClicked.connect(self.load_from_block)
        self.load_button = ui.memory_load_button
        self.load_button.clicked.connect(self.populate_hexdump)
        self.export_button = ui.memory_export_button
        self.export_button.clicked.connect(self.export)

    def show_with_data(self, address: int, size: int):
        """
        Displays given address in hexdump
        """
        self.start.setText(hex(address))
        self.size.setValue(size)
        self.populate_hexdump()
        self.ui.tabs.setCurrentWidget(self.ui.memory_tab)

    def _get_data(self) -> Optional[Tuple[int, bytes]]:
        """
        Obtain the start address and data currently requested in memory tab.
        """
        if not self._ctx:
            return

        start_address = self.start.text()
        if not start_address:
            # Clear hexdump if address is empty (startup)
            self.hexdump.setText("")
            return

        # First validate start address
        try:
            start_address = int(self.start.text(), 16)
        except ValueError as e:
            self.ui.show_exception(e, "Invalid address")
            return

        size = self.size.value()

        try:
            data = self._ctx.memory.read(start_address, size)
            return start_address, data
        except Exception as e:
            self.ui.show_exception(e, "Failed to read memory.")
            return

    def populate_hexdump(self):
        """
        Loads hexdump in memory tab.
        """
        ret = self._get_data()
        if not ret:
            return
        start_address, data = ret

        # Get hex dump
        lines = []
        for line in hexdump(data, result="generator"):
            # Replace address with one offset
            addr, _, rest = line.partition(":")
            addr = int(addr, 16) + start_address
            line = f"{addr:08X}: {rest}"
            lines.append(line)
        dump = "\n".join(lines)
        self.hexdump.setText(f"<pre>{dump}</pre>")

    def load_from_block(self, item: QtWidgets.QTableWidgetItem):
        if not self._ctx:
            return
        base_address, size = self._ctx.memory.blocks[item.row()]
        self.start.setText(hex(base_address))
        self.size.setValue(size)
        self.populate_hexdump()

    def export(self):
        """
        Exports current memory hexdump view to a file as raw binary.
        """
        if not self._ctx:
            return
        ret = self._get_data()
        if not ret:
            return
        start_address, data = ret

        file_path = ida_kernwin.ask_file(True, "*.bin", "Dump Memory")
        if file_path:
            with open(file_path, "wb") as outfile:
                outfile.write(data)

    def populate(self, ctx):
        """
        On initial load, show the memory map in the hex view.
        """
        self._ctx = ctx
        # Load the memory blocks table.
        blocks = ctx.memory.blocks
        table = self.blocks_table
        table.setRowCount(len(blocks))
        for index, (base_address, size) in enumerate(blocks):
            table.setItem(index, 0, QtWidgets.QTableWidgetItem(hex(base_address)))
            table.setItem(index, 1, QtWidgets.QTableWidgetItem(hex(base_address + size)))
            table.setItem(index, 2, QtWidgets.QTableWidgetItem(str(size)))

        # Load hexdump with any previously set fields.
        self.populate_hexdump()



"""
Utility for running YARA within the context of a disassembly.

This utility extends and overwrites the existing YARA API to work correctly within a dragodis context.

usage::
    from rugosa import yara

    # Compile a yara rule in the same way as the original yara library.
    rule = yara.compile(source=rule_text)

    with dragodis.open_program("input.exe") as dis:
        matches = rule.match(dis)  # Run rule on entire disassembled code.

        # Can also be used to run on segments.
        matches = rule.match(dis, segment='.text')

        # We can also just look for matching strings.
        for offset, identifer in rule.match_strings(dis):
            # ...
        for offset, identifer in rule.match_strings(dis, segment='.text'):
            # ...
"""

import logging
from typing import Iterable, List, Tuple, Union

import dragodis
from dragodis.interface import Function
import yara
from yara import *


logger = logging.getLogger(__name__)

READ_LENGTH = 10485760  # 10 MB


class Match:
    """
    Patches yara.Match to provide convert string offsets to virtual addresses.

    NOTE: We can't inherit yara.Match because they don't expose that class.

    :param yara.Match match_object: Original match object created by YARA
    :param int offset: Optional offset to offset string offsets by
    :param bool input_offset: Whether string offsets will be the file offset
        and should be converted.
    """

    def __init__(self, match_object, dis: dragodis.Disassembler, offset=None, file_offset=False):
        self._match = match_object
        self._dis = dis
        self._offset = offset
        self._file_offset = file_offset
        self._strings = None

    def __getattr__(self, item):
        return getattr(self._match, item)

    def __str__(self):
        return str(self._match)

    def __repr__(self):
        return repr(self._match)

    @property
    def strings(self):
        # Before returning strings, fixup the offsets to be virtual addresses.
        if self._strings is None:
            self._strings = []
            for offset, identifier, data in self._match.strings:
                if self._offset is not None:
                    offset += self._offset
                if self._file_offset:
                    offset = self._dis.get_virtual_address(offset)
                addr = self._dis.get_line(offset).address
                self._strings.append((addr, identifier, data))
        return self._strings


class Rules:
    """
    Patches yara.Rules to use our patched Match object when match() is called.

    NOTE: We can't inherit yara.Rules because they don't expose that class.
    """

    def __init__(self, rules_object):
        self._rules = rules_object
        self._infos = None

    def __getattr__(self, item):
        return getattr(self._rules, item)

    def _extract_info(self):
        """
        Retrieve information about the rule by performing a kludgy dance with callbacks.
        YARA should allow an easier way to give this information!
        """
        if self._infos is None:
            # YARA doesn't provide any easy way to get rule info, so we are going to have
            # to fake a match to get the info dictionary.
            self._infos = []

            def _callback(info):
                self._infos.append(info)
                return yara.CALLBACK_CONTINUE

            self._rules.match(data=b"", callback=_callback, which_callbacks=yara.CALLBACK_NON_MATCHES)
        return self._infos

    @property
    def names(self) -> List[str]:
        """Returns names of all the rules contained within."""
        infos = self._extract_info()
        return [info["rule"] for info in infos]

    def match(
            self, dis: dragodis.Disassembler, *args,
            input_offset=False, offset: int = None, segment: Union[str, int] = None, **kwargs
    ) -> List[Match]:
        """
        Patched to use our patched Match() object and allow for automatically running
        on IDB input file.

        Besides the default yara parameters, this implementation also includes:
            :param dis: Dragodis disassembler.
            :param *args: Positional arguments to pass to underlying yara.match() call.
            :param input_offset: Whether to apply input file offset to string offsets.
            :param offset: Optional offset to offset string offsets by.
            :param segment: Name or EA of segment to match to.
            :param **kwargs: Keyword arguments to pass to underlying yara.match() call.
        """
        # Run on segment.
        if segment:
            segment = dis.get_segment(segment)
            kwargs["data"] = segment.data
            offset = offset or segment.start
        # Run on input file.
        elif not (args or kwargs):
            args = (str(dis.input_path),)
            input_offset = True

        return [
            Match(match, dis, offset=offset, file_offset=input_offset)
            for match in self._rules.match(*args, **kwargs)
        ]

    def match_strings(self, *args, **kwargs) -> List[Tuple[int, str]]:
        """
        Runs match() but then returns tuples containing matched strings instead of Match objects.

        (This replicates the original legacy _YARA_MATCHES output)

        :returns: Tuple containing: (offset, identifier)
        """
        matched_strings = []
        for match in self.match(*args, **kwargs):
            for offset, identifier, _ in match.strings:
                matched_strings.append((offset, identifier))
        return matched_strings

    def find_functions(self, dis: dragodis.Disassembler, *args, **kwargs) -> Iterable[Function]:
        """
        Iterates functions that match the given rule text.
        """
        cache = set()
        for match in self.match(dis, *args, **kwargs):
            for ea, _, _ in match.strings:
                try:
                    func = dis.get_function(ea)
                except dragodis.NotExistError:
                    continue

                if func.start not in cache:
                    cache.add(func.start)
                    yield func


def compile(*args, **kwargs) -> Rules:
    """Wraps compiled rule in our patched Rules object."""
    return Rules(yara.compile(*args, **kwargs))


def load(*args, **kwargs) -> Rules:
    """Wraps loaded rule in our patched Rules object."""
    return Rules(yara.load(*args, **kwargs))


# Convenience functions ==============


def match(dis: dragodis.Disassembler, rule_text: str, *args, **kwargs) -> List[Match]:
    """Returns list of Match objects"""
    rule = compile(source=rule_text)
    return rule.match(dis, *args, **kwargs)


def match_strings(dis: dragodis.Disassembler, rule_text: str, *args, **kwargs) -> List[Tuple[int, str]]:
    """Returns list of (offset, string identifier)"""
    rule = compile(source=rule_text)
    return rule.match_strings(dis, *args, **kwargs)


def find_functions(dis: dragodis.Disassembler, rule_text: str, *args, **kwargs) -> Iterable[Function]:
    """
    Iterates functions that match the given rule text.
    """
    rule = compile(source=rule_text)
    yield from rule.find_functions(dis, *args, **kwargs)

# ====================================




from rugosa import re


def test_match(disassembler):
    ptn = re.compile(b"\x56\x56\x56\x56\x56\xe8\x9c\x12\x00\x00")
    match = ptn.search(disassembler)
    assert match
    assert match.start() == 0x4012c5
    funcs = list(re.find_functions(ptn, disassembler))
    assert len(funcs) == 1
    assert funcs[0].start == 0x4012a0



"""
Tests for bug fixes reported on GitHub.
"""

import pytest

import dragodis
from rugosa.emulation.emulator import Emulator
from rugosa.emulation.constants import WIDE_STRING


def test_issue_7(disassembler):
    """Tests the use of WIDE_STRING for read_data()"""
    emulator = Emulator(disassembler)
    context = emulator.new_context()

    wide_string = b"/\x00f\x00a\x00v\x00.\x00i\x00c\x00o\x00"
    context.memory.write(0x123000, wide_string)
    assert context.memory.read_data(0x123000, data_type=WIDE_STRING) == wide_string
    wide_string = b"\x00/\x00f\x00a\x00v\x00.\x00i\x00c\x00o"
    context.memory.write(0x123000, wide_string)
    assert context.memory.read_data(0x123000, data_type=WIDE_STRING) == wide_string


# TODO: Need to test this in both x86 and ARM
#   - Bring over or import the naming thing for disassembler.
def test_function_case_senstivity_all(disassembler):
    """Tests issue with case sensitivity when hooking functions."""
    emulator = Emulator(disassembler)

    # NOTE: Can't check equality of the functions directly since emulator could be teleported.

    # Test with known builtin func
    assert emulator.get_call_hook("lstrcpya").__name__ == "strcpy"
    assert emulator.get_call_hook("lStrcpyA").__name__ == "strcpy"
    assert emulator.get_call_hook("lstrcpyA").__name__ == "strcpy"

    # Test user defined
    def dummy(ctx, func_name, func_args):
        return

    assert emulator.get_call_hook("SuperFunc") is None
    assert emulator.get_call_hook("SUPERfunc") is None
    assert emulator.get_call_hook("superfunc") is None
    emulator.hook_call("SuperFunc", dummy)
    assert emulator.get_call_hook("SuperFunc").__name__ == "dummy"
    assert emulator.get_call_hook("SUPERfunc").__name__ == "dummy"
    assert emulator.get_call_hook("superfunc").__name__ == "dummy"


def test_multiple_emulations(strings_x86):
    """
    Tests issue with stale caching when multiple emulations happen in same process but different dragodis disassemblies.
    """
    address = 0x401024

    try:
        with dragodis.open_program(strings_x86, "ida") as dis:
            emulator = Emulator(dis, teleport=False)
            ctx = emulator.context_at(address)

        with dragodis.open_program(strings_x86, "ida") as dis:
            emulator = Emulator(dis, teleport=False)
            ctx = emulator.context_at(address)
    except dragodis.NotInstalledError as e:
        pytest.skip(str(e))

    # We pass if we don't get an EOFError raised.



from __future__ import annotations
from typing import TYPE_CHECKING

import jpype

RugosaActionListener = jpype.JClass("dc3.rugosa.plugin.RugosaActionListener")

from rugosa.ghidra_plugin.gui.error_msg import catch_errors
from rugosa.ghidra_plugin.gui.table import TableModel
from rugosa.ghidra_plugin.gui.context_menu import TableContextMenuEntry


if TYPE_CHECKING:
    from rugosa.ghidra_plugin.components.emulator import EmulatorForm


class ContextMenuEntry(TableContextMenuEntry):

    def __init__(self, variables: Variables):
        super().__init__(variables.table)
        self._variables = variables

    @property
    def variable(self):
        if self._variables._ctx:
            return self._variables._variables[self._component.getSelectedRow()]


class JumpToDisassembly(ContextMenuEntry):
    NAME = "Jump To Disassembly"

    def enabled(self, row: int, col: int) -> bool:
        return bool(self.variable)

    @catch_errors
    def clicked(self, row: int, col: int):
        self._variables.form.plugin.goTo(self._variables.form.dis._to_addr(self.variable.addr))


class ShowInMemory(ContextMenuEntry):
    NAME = "Show in Memory"

    @catch_errors
    def enabled(self, row: int, col: int) -> bool:
        var = self.variable
        return var and self._variables._ctx.memory.is_mapped(var.addr)

    @catch_errors
    def clicked(self, row: int, col: int):
        var = self.variable
        self._variables.form.memory.show_with_data(var.addr, var.size)


class Variables:

    MAX_VALUE_SIZE = 256
    HEADERS = ["Address", "Stack Offset", "Data Type", "Size", "Name", "Value"]

    def __init__(self, form: EmulatorForm):
        self._ctx = None
        self._variables = None
        self.form = form
        self.ui = form.ui
        self.table = self.ui.variablesTable
        self.form.add_context_menu_entry(JumpToDisassembly(self))
        self.form.add_context_menu_entry(ShowInMemory(self))

    @catch_errors
    def populate(self, ctx):
        """
        Populates the variables table.
        """
        self._ctx = ctx
        self._variables = list(ctx.variables)

        rows = []
        for var in self._variables:
            row = [""] * len(self.HEADERS)
            row[0] = hex(var.addr)
            if var.stack_offset is not None:
                row[1] = hex(var.stack_offset)

            data_type = var.data_type
            if var.count > 1 and data_type != "func_ptr":
                data_type += f"[{var.coune}]"
            row[2] = data_type

            row[3] = str(var.size)
            row[4] = var.name

            value = var.value
            if isinstance(value, (list, bytes)) and len(value) > self.MAX_VALUE_SIZE:
                value = str(value[:self.MAX_VALUE_SIZE]) + " (truncated)"
            else:
                value = str(value)
            row[5] = value

            rows.append(row)

        self.table.setModel(TableModel(self.HEADERS, rows))



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Tuple

import jpype
from hexdump import hexdump

RugosaActionListener = jpype.JClass("dc3.rugosa.plugin.RugosaActionListener")

from rugosa.ghidra_plugin.gui.error_msg import show_exception, catch_errors
from rugosa.ghidra_plugin.gui.table import TableModel
from rugosa.ghidra_plugin.gui.mouse_listener import MouseListener

if TYPE_CHECKING:
    from java.awt.event import ActionEvent, MouseEvent
    from rugosa.ghidra_plugin.components.emulator import EmulatorForm


class Memory:

    HEADERS = ["Start", "End", "Size"]

    def __init__(self, form: EmulatorForm):
        self._ctx = None
        self.form = form
        self.ui = form.ui
        self.start = self.ui.memoryStart
        self.size = self.ui.memorySize
        self.hexdump = self.ui.memoryHexDump
        self.blocks_table = self.ui.memoryBlocksTable

        self.blocks_table.addMouseListener(MouseListener(clicked=self.load_from_block))
        self.ui.memoryLoadButton.addActionListener(RugosaActionListener @ (lambda event: self.populate_hexdump()))
        self.ui.memoryExportButton.addActionListener(RugosaActionListener @ self.export)

    @catch_errors
    def show_with_data(self, address: int, size: int):
        """
        Displays given address in hexdump
        """
        self.start.setText(hex(address))
        self.size.setValue(size)
        self.populate_hexdump()
        self.ui.tabs.setSelectedComponent(self.ui.memoryTab)

    def _get_data(self) -> Optional[Tuple[int, bytes]]:
        """
        Obtain the start address and data currently requested in memory tab.
        """
        if not self._ctx:
            return

        start_address = self.start.getText()
        if not start_address:
            # Clear hexdump if address is empty (startup)
            self.hexdump.setText("")
            return

        # First validate start address
        try:
            start_address = int(self.start.getText(), 16)
        except ValueError as e:
            show_exception(e, "Invalid address")
            return

        size = self.size.getValue()

        try:
            data = self._ctx.memory.read(start_address, size)
            return start_address, data
        except Exception as e:
            show_exception(e, "Failed to read memory.")
            return

    @catch_errors
    def populate_hexdump(self):
        """
        Loads hexdump in memory tab.
        """
        ret = self._get_data()
        if not ret:
            return
        start_address, data = ret

        # Get hex dump
        lines = []
        for line in hexdump(data, result="generator"):
            # Replace address with one offset
            addr, _, rest = line.partition(":")
            addr = int(addr, 16) + start_address
            line = f"{addr:08X}: {rest}"
            lines.append(line)
        dump = "\n".join(lines)
        self.hexdump.setText(dump)
        self.hexdump.scrollToTop()

    @catch_errors
    def load_from_block(self, event: MouseEvent):
        if not self._ctx:
            return
        if event.getClickCount() == 2:  # double click
            base_address, size = self._ctx.memory.blocks[event.getSource().getSelectedRow()]
            self.start.setText(hex(base_address))
            self.size.setValue(size)
            self.populate_hexdump()

    @catch_errors
    def export(self, event: ActionEvent):
        """
        Exports current memory hexdump view to a file as raw binary.
        """
        if not self._ctx:
            return
        ret = self._get_data()
        if not ret:
            return
        start_address, data = ret

        from docking.widgets.filechooser import GhidraFileChooser
        chooser = GhidraFileChooser(self.ui.memoryTab)
        chooser.setMultiSelectionEnabled(False)
        chooser.setTitle("Dump Memory")
        file = chooser.getSelectedFile()
        if file:
            with open(file.getPath(), "wb") as outfile:
                outfile.write(data)

    @catch_errors
    def populate(self, ctx):
        """
        Populates the variables table.
        """
        self._ctx = ctx
        # Load the memory blocks table.
        rows = []
        for base_address, size in ctx.memory.blocks:
            rows.append([hex(base_address), hex(base_address + size), str(size)])
        self.blocks_table.setModel(TableModel(self.HEADERS, rows))

        # Load hexdump with any previously set fields.
        self.populate_hexdump()
